#!/usr/bin/env python
# encoding: utf-8
"""
$Id:$

Purpose: Run the ADL 3.1 CRIS SDR

Input:
    One or more HDF5 CRIS RDR input files, aggregated or single-granule.
    You need four granules before and four granules after a given granule to process.
    A work directory, typically, empty, in which to unpack the granules and generate the output.
    If the work directory specified does not exist, it will be created.

Output:
    ADL CRIS SDR blob files, .lutasc metadata files, and HDF5 output granules will be created.

Details:
    If you input a series of granules, the software will scan the work directory.
    Thus, for N input RDR granules, you may get up to N-8 output SDR granules if they are all contiguous.
    It is ambiguous to provide several copies of the same granule in the work directory;
    this will result in an error abort.
    The unpacker gives each unpacking of a given granule its own

    Repairs .lutasc files if necessary to correct the following issues found in output
    from RTPTP 5.0:
       - n_collection_short_name 'CrIS-SCIENCE-RDR' (should be 'CRIS-SCIENCE-RDR')
       - Invalid granule IDs (times are out of range tolerated by ADL)

Preconditions:
    Requires ADL_HOME is set.
    Requires that any needed LD_LIBRARY_PATH is set.

Created 2011 G Martin <gmartin@ssec.wisc.edu>
Copyright (C) 2011-2013 University of Wisconsin Regents.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys
import logging
import glob
import traceback
import shutil
import time
import datetime as dt
from subprocess import CalledProcessError, call
from shutil import rmtree
from sdr_process import Sdr_process
#from sdr_process import add_geo_attribute_to_h5
# skim and convert routines for reading .lutasc metadata fields of interest

from adl_common import  ADL_HOME,  sh
from sdr_config_paths import get_and_check_cspp_dynamic_ancil_dir
from adl_common import env
from sdr_config_paths import check_installation

from adl_common import   h5_xdr_inventory, get_return_code, check_env
from cspp_some_utils.log_common import configure_logging, status_line

from adl_common import CSPP_RT_HOME, initialize_work_ADL_variables, ADL_VARS

from adl_query_dms import  incoming_luts_abs_path
from adl_common import cspp_x_home  


from cspp_some_utils.basics import check_and_convert_path
from cspp_some_utils.basics import check_existing_env_var
from cspp_some_utils.basics import check_and_convert_env_var

from adl_cris_data import  ADL_CRIS_SDR 

from adl_cris_data import  CRIS_SDR_PRODUCTS_AGG,CRIS_SDR_PRODUCTS_AGG_GEO
from adl_cris_data import  FS_CRIS_SDR_PRODUCTS_AGG,FS_CRIS_SDR_PRODUCTS_AGG_GEO
from adl_cris_data import  XML_TMPL_CRIS_SDR,ADL_CRIS_FS_SDR
from adl_cris_data import  XML_TMPL_CRIS_FS_SDR, extra_prod

# cache loading and searching
import rdr_unpack
import datetime
# N_GEO_Ref fix for ADL
# from adl_geo_ref import write_geo_ref

from sdr_pre_process import Sdr_pre_process 

from multiprocessing import Pool

from adl_cris_data import GIG_MEMORY_PER_PROC, CRIS_GRAN_ID_DELTA

#from granule_status import Granule_processing_status    
from adl_log import checkADLLogForSuccess    
from system_checks import System_checks

from adl_query_dms import  setup_dms_in_work_directory
import adl_anc_retrieval
from ancillary_control import check_for_cache_database_initalization
import pprint
pp = pprint.PrettyPrinter(indent=4)

# every module should have a LOG object
LOG = logging.getLogger(__name__)
#granule_status = None

CALWINDOW='normal'
from adl_sdr import ADL_SDR

class adl_cris_sdr(ADL_SDR):
    """
    """
    
    fullresolution=True
    def __init__(cls,work_dir,allow_cache_update,instrument,spacecraft,fullres):
        """
        """
        super(adl_cris_sdr,cls ).__init__(work_dir,allow_cache_update,instrument,spacecraft)

        if fullres is False :  
            cls.fullresolution = False
            cls.sdr_products=CRIS_SDR_PRODUCTS_AGG
            cls.agg_products=CRIS_SDR_PRODUCTS_AGG
            cls.geo_products=CRIS_SDR_PRODUCTS_AGG_GEO 
            cls.controllers.append(ADL_CRIS_SDR)           
            cls.controllers_xml.append(XML_TMPL_CRIS_SDR) 
            fnxml = 'sdr_cris_%s.xml'  
        else :
            cls.sdr_products=FS_CRIS_SDR_PRODUCTS_AGG
            cls.agg_products=FS_CRIS_SDR_PRODUCTS_AGG
            cls.geo_products=FS_CRIS_SDR_PRODUCTS_AGG_GEO 
            cls.controllers.append(ADL_CRIS_FS_SDR)           
            cls.controllers_xml.append(XML_TMPL_CRIS_FS_SDR)  
            fnxml = 'sdr_cris_fs_%s.xml'

        cls.fnxmls.append(fnxml)
              
        cls.expected_granule_delta = CRIS_GRAN_ID_DELTA   
        cls.GIG_MEMORY_PER_PROC = GIG_MEMORY_PER_PROC
        
    def check_cmo(cls,argument_dictionaries,granules_to_process):
        """
        """
        # the first granule needs to be processed before multi-processing so
        # that the CMO gets generated only once.
        # in pipelined operation the CMO may already exits or be in the process
        # of getting created
        # if the lutasc file for the CMO exists the CMO creation is complete, if the directory for the
        # first granule exists does the MO does not we must wait.
    
        if len (granules_to_process) <=  0:
            LOG.info("No granules to process")
            return argument_dictionaries 
        gran = granules_to_process[0]
        gran_id = gran['N_Granule_ID']
        granule_output_dir = os.path.join(cls.work_dir, gran_id)
        
        run_cmo = cls.run_cmo
        noncritical_problem = 0
        
        if not os.path.isdir(granule_output_dir):
            run_cmo = True
        
        # already present in work directory - user supplied
        if cls.fullresolution is True:
            cmo_path = os.path.join(cls.work_dir, "CrIS-FS-Correct-Matrix*.h5")
        else:
            cmo_path = os.path.join(cls.work_dir, "CrIS-Correct-Matrix*.h5")
          
        tmplate = cls.controllers_xml[0]
        fnxml = cls.fnxmls[0]

        files = glob.glob(cmo_path)
        if len(files) >= 1:
            run_cmo = False
            
        first_granule_cleanup = cls.cleanup
        created_cmo = False
        cmo_attempted = False
        #run_cmo = False      
        
        while run_cmo is True:
            cmo_attempted = True
            # generate CMO
            if len(argument_dictionaries) > 0:
                cmo_env = argument_dictionaries.pop(0)
                LOG.info(
                    "Check/Generate CMO with Granule %s.", cmo_env['N_Granule_ID'])
                result = cls.run_and_check_single_granule(cmo_env)

                cmo_files = glob.glob(cmo_path)
                if len(cmo_files) >= 1:
                    LOG.info('CMO generated %s' % cmo_files[0])
                    run_cmo = False
                    result['GOOD'] = "True"

                if result['GOOD'] == "True":
                    run_cmo = False
                    #2018_01_22_022
                    if len(cmo_files) >= 1:
                        cmo_file=os.path.basename(cmo_files[0])
                        #cmo_time=cmo_file.split('_')[3][0:8]
                        
                        #desired_dt = datetime.datetime.strptime(cmo_time , "%Y%m%d")
                        #anc_string = desired_dt.strftime('%Y_%m_%d_%j')
                        cmo_file="CrIS*.h5"
                        local_directory_for_luts = incoming_luts_abs_path()
                        #status_line('mkdir -p ${CSPP_SDR_HOME}/anc/cache/%s;cp %s ${CSPP_SDR_HOME}/anc/cache/%s '%(anc_string,cmo_file,anc_string) )
                        status_line('mv %s %s '%(cmo_file,local_directory_for_luts) )
                        created_cmo = True
                        try:
                            #dest=os.path.join(cspp_x_home, 'anc', 'cache',anc_string)
                            os.mkdir(local_directory_for_luts)
                        except :
                            pass
                        try :
                            shutil.move(cmo_files[0], local_directory_for_luts)
                            LOG.info('CMO moved %s %s'%(cmo_files[0], local_directory_for_luts))
                        except :
                            pass
                
                    if cls.cleanup is True:
                        granule_dir = cmo_env['GRANULE_OUTPUT_DIR']
                        LOG.debug("Remove: " + granule_dir)
                        LOG.debug("Remove: " + granule_dir)
                        if os.path.exists(granule_dir):
                            rmtree(granule_dir)
                else:
                    LOG.error('Problem with CMO generation granule. %s' %
                            cmo_env['N_Granule_ID'])
                    if first_granule_cleanup is True:
                        granule_dir = cmo_env['GRANULE_OUTPUT_DIR']
                        LOG.debug("Remove: " + granule_dir)
                        LOG.debug("Remove: " + granule_dir)
                        if os.path.exists(granule_dir):
                            rmtree(granule_dir)
                        first_granule_cleanup = False
    
                    noncritical_problem += 1

            else:
                LOG.error("No CMO generated")
                run_cmo = False
        
        cmo_path = os.path.join(cls.work_dir, "*Matrix-AUX*.h5")
        cmo_files = glob.glob(cmo_path)
        
        cmo_usage=str() 
        if cmo_attempted is True : 
            cmo_usage += "No ancillary CMO provided, "
            if created_cmo is True:
                cmo_usage += "generated by software.  "
            else :
                cmo_usage += "CMO generation failed  "

        else :
            cmo_usage += "Ancillary CMO provided. "  
            
        if len(cmo_files) >= 1: 
            cmo_usage.append(" %s"%cmo_files[0])          
                
        status_line('CMO usage: %s'%cmo_usage)
        
        return argument_dictionaries
           
# def run_all_granules(work_dir, 
#                      granules_to_process, 
#                      trim=None,  
#                      cleanup=True, 
#                      compress=False,
#                      nprocs="1",
#                      spacecraft='NPP', 
#                      no_chunk=False,
#                      accuracyMode='reduced',
#                      INSTRUMENT='CRIS',
#                      fullresolution=False,
#                      extra=False,
#                      **additional_env):
#     """run each CRIS SDR XML input in sequence
#     return info regarding problems encountered: number of bad logs,
#     """
#     global granule_status
#     if not trim:
#         trim = [4, 0]

#     global total_granules
#     noncritical_problem = 0

#     total_granules = len(granules_to_process)
#     if len(granules_to_process) == 0:
#         return 1

#     gran = granules_to_process[0]
#     gran_id = gran['N_Granule_ID']
#     granule_output_dir = os.path.join(work_dir, gran_id)

#     if not os.path.isdir(granule_output_dir):
#         run_cmo = True
        
#     run_cmo =  not extra
 
#     # already present in work directory - user supplied
#     if fullresolution is True:
#         fnxml = 'sdr_cris_%s.xml'
#         tmplate = XML_TMPL_CRIS_FS_SDR
#         cmo_path = os.path.join(work_dir, "CrIS-FS-Correct-Matrix*.h5")

#     else:
#         fnxml = 'sdr_cris_%s.xml'
#         tmplate = XML_TMPL_CRIS_SDR
#         cmo_path = os.path.join(work_dir, "CrIS-Correct-Matrix*.h5")

#     files = glob.glob(cmo_path)
#     if len(files) >= 1:
#         run_cmo = False
        
#     #sdr_pp = Sdr_pre_process(work_dir,"CRIS",spacecraft)   
#     LOG.debug('Run CMO %s'%str(run_cmo))
#     #granules_to_process = sdr_pp.trim_the_granules(
#     #    granules_to_process, CRIS_GRAN_ID_DELTA, trim=trim)

#     inproc= int( nprocs )
#     fudge = inproc
#     if int(  len(granules_to_process) % inproc ) == 0 :
#         fudge = 0
    
#     chunk_size= int(  (len(granules_to_process) + fudge )  / inproc )
#     if chunk_size <= 0:
#         chunk_size=1
    
#     if no_chunk is True:
#         chunk_size = 1

#     additional_keys = dict(spacecraft=spacecraft, 
#                            FULLRES=str(fullresolution), 
#                            CHUNK_SIZE=str(chunk_size),
#                            INSTRUMENT=INSTRUMENT,
#                            CALWINDOW=CALWINDOW)

  
#     sdr_proc = Sdr_process(work_dir,INSTRUMENT,spacecraft)

#     argument_dictionaries, granules_to_process = sdr_proc.setup_multi_process(
#                                                                      granules_to_process,
#                                                                      [fnxml],
#                                                                      [tmplate],
#                                                                      additional_keys,
#                                                                      accuracyMode=accuracyMode)

#     total_granules = len(granules_to_process)
#     granule_status = Granule_processing_status(total_granules)

#     # the first granule needs to be processed before multi-processing so
#     # that the CMO gets generated only once.
#     # in pipelined operation the CMO may already exits or be in the process
#     # of getting created
#     # if the lutasc file for the CMO exists the CMO creation is complete, if the directory for the
#     # first granule exists does the MO does not we must wait.

#     first_granule_cleanup = cleanup
#     created_cmo = False
#     while run_cmo is True:
#         # generate CMO
#         if len(argument_dictionaries) > 0:
#             cmo_env = argument_dictionaries.pop(0)
#             LOG.info(
#                 "Check/Generate CMO with Granule %s.", cmo_env['N_Granule_ID'])
#             result = run_and_check_single_granule(cmo_env)

#             cmo_files = glob.glob(cmo_path)
#             if len(cmo_files) >= 1:
#                 LOG.info('CMO generated %s' % cmo_files[0])
#                 run_cmo = False
#                 result['GOOD'] = "True"

#             if result['GOOD'] == "True":
#                 run_cmo = False
#                 #2018_01_22_022
#                 if len(cmo_files) >= 1:
#                     cmo_file=os.path.basename(cmo_files[0])
#                     #cmo_time=cmo_file.split('_')[3][0:8]
                    
#                     #desired_dt = datetime.datetime.strptime(cmo_time , "%Y%m%d")
#                     #anc_string = desired_dt.strftime('%Y_%m_%d_%j')
#                     cmo_file="CrIS*.h5"
                    
#                     local_directory_for_luts = incoming_luts_abs_path()

                    
#                     #status_line('mkdir -p ${CSPP_SDR_HOME}/anc/cache/%s;cp %s ${CSPP_SDR_HOME}/anc/cache/%s '%(anc_string,cmo_file,anc_string) )
#                     status_line('mv %s %s '%(cmo_file,local_directory_for_luts) )
#                     created_cmo = True
#                     try:
#                         #dest=os.path.join(cspp_x_home, 'anc', 'cache',anc_string)
#                         os.mkdir(local_directory_for_luts)
#                     except :
#                         pass
#                     try :
#                         shutil.move(cmo_files[0], local_directory_for_luts)
#                         LOG.info('CMO moved %s %s'%(cmo_files[0], local_directory_for_luts))
#                     except :
#                         pass
            

#                 if cleanup is True:
#                     granule_dir = cmo_env['GRANULE_OUTPUT_DIR']
#                     LOG.debug("Remove: " + granule_dir)
#                     LOG.debug("Remove: " + granule_dir)
#                     if os.path.exists(granule_dir):
#                         rmtree(granule_dir)
#             else:
#                 LOG.error('Problem with CMO generation granule. %s' %
#                           cmo_env['N_Granule_ID'])
#                 if first_granule_cleanup is True:
#                     granule_dir = cmo_env['GRANULE_OUTPUT_DIR']
#                     LOG.debug("Remove: " + granule_dir)
#                     LOG.debug("Remove: " + granule_dir)
#                     if os.path.exists(granule_dir):
#                         rmtree(granule_dir)
#                     first_granule_cleanup = False
 
#                 noncritical_problem += 1

#         else:
#             LOG.error("No CMO generated")
#             run_cmo = False
     


#     cmo_path = os.path.join(work_dir, "*Matrix-AUX*.h5")
#     cmo_files = glob.glob(cmo_path)
    
#     cmo_usage=str() 
#     if extra is False : 
#         cmo_usage += "No ancillary CMO provided, "
#         if created_cmo is True:
#             cmo_usage += "generated by software.  "
#         else :
#             cmo_usage += "CMO generation failed  "

#     else :
#         cmo_usage += "Ancillary CMO provided. "  
        
#     if len(cmo_files) >= 1: 
#         cmo_usage.append(" %s"%cmo_files[0])          
            
#     status_line('CMO usage: %s'%cmo_usage)
    

#     keep_all_output = False
#     results = []
#     # results = map (run_and_check_single_viirs_sdr, argument_dictionaries )
#     # if there are still granules to process and a cmo has been generated.
#     # if len(argument_dictionaries) > 0 and len(cmo_files) > 0:
#     if len(argument_dictionaries) > 0:

#         sys_checks = System_checks(int(nprocs),GIG_MEMORY_PER_PROC)
#         nprocs = sys_checks.number_processors_used

#         parallel = Pool(int(nprocs))

#         try:
#             results = parallel.map(
#                 run_and_check_single_granule, 
#                 argument_dictionaries)
#             #results = parallel.map_async(
#             #    run_and_check_single_granule, 
#             #    argument_dictionaries).get(9999999)
            
#         except KeyboardInterrupt:
#             LOG.error("Got exception, stopping workers and exiting.")
#             parallel.terminate()
#             parallel.join()
#             sys.exit(1)




#     return noncritical_problem

def main():
    import argparse

    desc = """
Run the CrIS SDR algorithm on the specified HDF5 RDR file(s). Output SDR and
Geo files will go in the specified working directory (defaults to current
directory)
"""
    parser = argparse.ArgumentParser(description=desc)
    #    parser.add_argument('-t', '--test',
    #                action="store_true", default=False, help="run self-tests")
    parser.add_argument('-l', '--local',
                        action="store_true", default=False, help="disable download of remote ancillary data to cache")
    parser.add_argument('-W', '--work-dir', metavar='work_dir', default='.',
                        help='directory in which all activity occurs, defaults to current dir')
    parser.add_argument('-v', '--verbosity', action="count", default=0,
                        help='each occurrence increases verbosity 1 level through ERROR-WARNING-INFO-DEBUG')
    parser.add_argument('-c', '--cris-cmo-dir', metavar='cris_cmo_dir', default='',
                        help='optional, directory containing .bin CrIS CMO file(s)')
    parser.add_argument('-d', '--debug',
                        action="store_true", default=False, help="always retain intermediate files")
    parser.add_argument('-t', '--trim', nargs=2, type=int, metavar='N',
                        default=None,
                        help='Number Trim granules. -t \"Trim Begin\" \"Trim End\" -t 1 1 Trims first and last granule.')

    parser.add_argument('-s', '--no-size-threshold',
                        action="store_true", default=False,
                        help="do not apply size threshold to RDRs. Use at your own risk; may cause ADL to crash or hang.")

    parser.add_argument('-z', '--zip',
                        action="store_true", default=False, help="compress products with h5repack zip compression")

    parser.add_argument('-f', '--fsr',
                        action="store_true", default=None, help="FSR res product")

    parser.add_argument('-n', '--nsr',
                        action="store_true", default=None, help="NSR product")

    parser.add_argument('-a', '--aggregate',
                        action="store_true", default=False, help="aggregate products with nagg")
    
#     parser.add_argument('-m', '--accuracymode',
#                         action="store_true", default=False, help="increase geo accuracy")

    parser.add_argument('--nochunk',
                        action="store_true", default=False, help="Slow mode")
 
    parser.add_argument('--center', dest='center', action='store_true',
                        help='Trim to center granule')
    parser.add_argument('--no-center', dest='center', action='store_false',
                        help='Normal output')
    parser.set_defaults(center=False)

    parser.add_argument('-p', '--processor', metavar='n',
                        type=int, default=1, help="Number of processors to use for granule processing, defaults to 1")
    parser.add_argument('filenames', metavar='filename', type=str, nargs='+',
                        help='HDF5 CrIS RDR file to process')
    
    
    

    args = parser.parse_args()

    levels = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
    level = levels[args.verbosity if args.verbosity < 4 else 3]

    work_dir = check_and_convert_path(args.work_dir,create=True,key="WORK_DIR")
    d = dt.datetime.now()
    timestamp = d.isoformat()
    logname = "cris_sdr." + timestamp + ".log"
    logfile = os.path.join(work_dir, logname)
    configure_logging(level, FILE=logfile)
    docleanup = True
    if args.debug is True:
        docleanup = False  
    
    LOG.debug('verbosity level is %d' % args.verbosity)

    compress = args.zip
    aggregate = args.aggregate

    global CALWINDOW
    accuracyMode='reduced'
#     if args.accuracymode is True:
#          accuracyMode = 'full'

    CALWINDOW='normal'
#     if args.calwindow is True:
#          CALWINDOW = 'reset'

    
    work_dir = os.path.abspath(args.work_dir)
    cris_cmo_dir = args.cris_cmo_dir
    if cris_cmo_dir:
        cris_cmo_dir = os.path.abspath(cris_cmo_dir)
    if not args.trim:
        args.trim = [4, 4]

    if not args:
        parser.print_help()
        return 9
    
    dpe_version = check_existing_env_var('DPE_VER', default_value=None, silent=True)
    LOG.info("product version: DPE_VER for SDR  %s" % dpe_version)
    check_installation(CSPP_RT_HOME)
    LOG.info('Arguments \n %s ' % pp.pformat(args))
    work_dir = os.path.abspath(args.work_dir)
    LOG.info('CSPP execution work directory is %r' % work_dir)
    
    get_and_check_cspp_dynamic_ancil_dir(check_write=not args.local)
    check_for_cache_database_initalization(CSPP_RT_HOME,work_dir,allow_cache_update=not args.local)

    full_db_name = setup_dms_in_work_directory(work_dir, CSPP_RT_HOME)
    initialize_work_ADL_variables(work_dir, full_db_name)

    check_env()

    rc = 1
    try:
        
        fullres = True
        if args.nsr is not None:
            fullres = False
        
        if 'npp' in os.path.basename( args.filenames[0] ) :
            spacecraft = 'NPP'
        elif  'j01' in os.path.basename( args.filenames[0] ) :
            spacecraft = 'J01'
        elif  'j02' in os.path.basename( args.filenames[0] ) :
            spacecraft = 'J02'
        else :
            spacecraft = "BAD"

        filenames = args.filenames
        
        instrument=None
        if fullres is True:
            instrument='CRIS'
        else :
            instrument='CRIS'
         
        allow_cache_update=not args.local   
        comps_sdr=adl_cris_sdr(work_dir,allow_cache_update,instrument,spacecraft,fullres)

        comps_sdr.proc_cfg(args.trim,args.processor) 
        comps_sdr.devel_cfg(accuracyMode,docleanup,args.center,not args.no_size_threshold)   
        comps_sdr.product_cfg(args.aggregate,args.zip)


        rc = comps_sdr.sdr(filenames)

        if not args.debug:
            os.remove(logfile)
            comps_sdr.cleanup_work_dir(force_all=True)
  
    except Exception:
        LOG.error(traceback.format_exc())

    return rc


if __name__ == '__main__':
    sys.exit(main())
